#include "cspec.h"
/* ===============================================================

   File genbc.cfn, July 2003:   Contains function genbc() used
   to generate coefficients b[], c[] for data-point lookups.

   Explanation:  Suppose we have N samples of a function f(x) at
   points x = x0, x0+dx0, ... x0+(N-1)*dx0.   We wish to 
   estimate f(x) for arbitrary x within that range.  We'll need 
   the following stored quantities: 

   -- Three data arrays a[i], b[i], c[i], described below;  
   -- x0 and dx0;
   -- in order to avoid divide operations, we also keep 
        another parameter,  recdx0 = 1./dx0.

   Then, for a given value of x, here's the procedure:  

   (1) Calculate  w = (x - x0)*recdx0.   

   (2) ww = w + 0.5.  

   (3) Convert ww to integer n.  (Truncate the fraction.)
       This identifies the lookup point that's closest to x;  
       shortly we'll use stored quantities a[n], b[n], and c[n].

   (4) Generally it's prudent to verify that 
                    0 <= n <= N-1.
       If n isn't in that range, then refuse to estimate f(x), or
       maybe assign some standard error-value to it.    

       In some cases, however, we may know that x is definitely is 
       in the range  x0 < x < x0 + (N-1)*dx0.  Omitting the range
       test might speed the process up because "if" comparisons 
       may be slower than arithmetic opertions.   
        
   (5) Calculate  s = w - n.   [Caution:  w - n, not ww - n.]  
       Note that s will be in the range -0.5 to +0.5. 
       [Alternatively s = (fractional part of ww) - 0.5.]

   (6) Our estimated value of f(x) is  a[n] + s*(a[b] + s*c[n]).

       [This is just a + b*s + c*s^2, evaluated efficiently.
       Evidently a[n] must be the functional value f(n),
       b[n] is the derivative there, and c[n] is half the
       second derivative.]

    In summary, we require about 9 floating-point arithmetic 
    operations to calculate f(x), not including "if" comparisons 
    to verify that x is within the legal limits.  If necessary
    we could reduce the 9 to about 7 f.p. operations, by
    using slightly different versions of arrays a[], b[], c[]. 

    We must prepare data arrays a[], b[], c[] beforehand. 
    First we just read a set of data values a[n] = uniformly-spaced 
    samples of f(x).  [When doing that, we also note N, the number 
    of points in the list.]   After N and a[] are safely stored, 
    then we generate b[] and c[] by calling function genbc(N,a,b,c), 
    which is copied below. 

    [Note that genbc() includes three mini-functions.  Also,
    it normally does NOT need #include <math.h>.] 

  =============================================================== */ 


/* Function genbc(), July 2003:  Generates coefficients B, C for
   fast interpolation.

   Suppose we wish to estimate values of a function f(x) based
   on a list of uniformly spaced data points f[i].  Use the
   following interpolation procedure.   First identify the integer
   i that's nearest to x:   i-0.5 <= x <= i+0.5.   Then calculate
   s = x - i  (-0.5 <= s <= +0.5),  and the desired function value
   is approximately given by

           f(x) = a[i] + b[i]*s + c[i]*s^2
                = a[i] + s*(b[i] + s*c[i]).

   Coefficient a[i] is simply f[i], and the following routine
   generates the b[i] and c[i] coefficients, which are essentially
   the local first and second derivatives.

   Input parameter npts specifies the number of data points
   stored in array a[i]:  i = 0 ... npts-1.   If npts < 1,
   then this program does nothing.  Of course the arrays
   a[], b[], c[] must have been declared to contain at least 
   npts elements.   This routine does not use or store any 
   elements outside the range i = 0 ... npts-1. 

   In some cases we might wish to declare a[], b[], c[]
   as floats rather than doubles  (fp4 instead of fp8).

   Note that we use mini-functions cformula(), terpol6(), 
   and terpol4(), included below.  

   ----------------------------------------------------------- */

void genbc(int npts, double * a, double * b, double * c)
 {

  double cformula(double, double, double);
  double terpol6(int, double *), terpol4(int, double *);
  int m;

  double f_a,f_b;
  double zero=0.,half=0.5;

  
  /* The main part of this program is short and easy, the m-loop below.
     If npts > 6, this m-loop calculates the "central set" of b[m], c[m] 
     values for m = 3 to npts-4.  Each calculation is based on 7 data 
     points a[m-3] ... a[m+3], using a technique explained in comments 
     below.   Later we'll use different formulae to estimate values near 
     the ends, for m < 3 and m > npts-4. */   

  if (npts > 6)   
   {
    f_b = terpol6(2,a);   /* to initialize f_a, f_b; explained below */
    for (m=3;m<=npts-4;m++)  
     {
      f_a = f_b;
      f_b = terpol6(m,a); 
      /*  f_a and f_b are interpolated values at x = m-0.5 and m+0.5.  
          Therefore, now we have three local sample values f(m-0.5),
          f(m), and f(m+0.5).  We fit a quadratic f(s) to them, where
          s = x - m  so  s  is between -0.5 and +0.5.  For each point 
          the first coefficient is simply a[m], already known.  */ 
      b[m] = f_b - f_a;  
      c[m] = cformula(f_a,a[m],f_b);   
     }
   }     /* ---  end of "if (npts > 6)"  --- */


  /* Now we estimate values near the ends,  m < 3 or m > npts-4.
     These few points unfortunately occupy most of the program, 
     because small data sets with npts < 7 require varying treatments.  
     Here we use more "if" statements than are strictly needed, 
     in order to maximize clarity.   Note:  If npts < 1, then this 
     program does nothing.  */ 
  
  if (npts == 1) { b[0] = zero;  c[0] = zero; } 

  if (npts == 2)
    { b[0] = a[1] - a[0];  b[1] = b[0];  c[0] = zero;  c[1] = zero; } 

  if (npts > 2)
   {
      /* Use a simple 3-pt quadratic fit for the end points
         [0] and [npts-1]: */  
    c[0] = 0.25*cformula(a[0],a[1],a[2]);  /* simple quadratic fit */ 
    b[0] = a[1] - a[0] - c[0]; 
    m = npts - 1;
    c[m] = 0.25*cformula(a[m-2],a[m-1],a[m]);
    b[m] = a[m] - a[m-1] + c[m]; 
      /* The next line merely provides for the case npts = 3;  
         if npts > 3, then these values of b[1] and c[1] will 
         soon be replaced. */ 
    c[1] = c[0];  b[1] = half*(a[2] - a[0]);  
   }  
 
     /* At this point, b and c have been calculated for the end 
        points [0] and [npts-1],  and b[1] and c[1] have been 
        calculated for the special case npts = 3.  */ 

  if (npts > 3)   /* Calculate next-to-end points */
   {
    f_a = a[0] + half*(b[0] + half*c[0]);   /* f(0.5) */
    f_b = terpol4(1,a);                     /* f(1.5) */
    c[1] = cformula(f_a,a[1],f_b); 
    b[1] = f_b - f_a;  
    m = npts - 1;
    f_a = terpol4(m-2,a);                   /* f(npts-2.5) */
    f_b = a[m] - half*(b[m] - half*c[m]);   /* f(npts-1.5) */
    --m;   
    c[m] = cformula(f_a,a[m],f_b);         /* m = npts - 2 */ 
    b[m] = f_b - f_a;
   }  
          
  if (npts == 5)  /* Calculate point [2] for this special case */  
   {
    f_a = terpol4(1,a);  /* f(1.5) */
    f_b = terpol4(2,a);  /* f(2.5) */
    c[2] = cformula(f_a,a[2],f_b); 
    b[2] = f_b - f_a;
   }

  if (npts > 5)  /* Calculate 2nd-from-end points */
   {
    f_a = terpol4(1,a);                 /* f(1.5) */
    f_b = terpol6(2,a);                 /* f(2.5) */
    c[2] = cformula(f_a,a[2],f_b);
    b[2] = f_b - f_a; 
    m = npts - 4;
    f_a = terpol6(m,a);               /* f(npts-3.5) */ 
    ++m;
    f_b = terpol4(m,a);               /* f(npts-2.5) */
    c[m] = cformula(f_a,a[m],f_b);    /* m = npts - 3 */
    b[m] = f_b - f_a;
   }

 }  /*  ---------- end of function genbc() ----------  */


/* ----------------  Three mini-functions: --------------------- */ 

/* cformula():  If the three arguments are f(x0-0.5), f(x0), and
   f(x0+0.5), this formula gives the local quadratic coefficient
   C in  f(x) = A + B*(x-x0) + C*(x-x0)^2  -- or, in other words,
   C = half the second derivative at x = x0. 
   --------------------------------------------------- */
double cformula(double fa, double f0, double fb) 
 {
  return 2.*(fa + fb - 2.*f0);
 }  /*  ----- end of cformula()  ----- */


/* ----------------------------------------------------------------
   Functions terpol4(m,f) and terpol6(m,f):   Estimate f(m+0.5) by
   4-point or 6-point interpolation, respectively.  These simple 
   formula occur frequently in the above routine genbc().  */

double terpol6(int m, double * f) 
 {
  double q,r6[3]={ 0.5945, -0.1105, 0.0160 }; 
  /* These coefficients are optimized for Fourier-component wavelengths
     near 5 pixels.   Then also give exact results for  f(x) = x^2.  */ 
  q = r6[0]*(f[m]+f[m+1]) + r6[1]*(f[m-1]+f[m+2]) + r6[2]*(f[m-2]+f[m+3]);
  return q;
 }

/* terpol4():  4-point interpolation to find value f(m+0.5). */ 
double terpol4(int m, double * f) 
 {
  /* These coefficients give exact results for f(x) = x^2. */
  return 0.5625*(f[m] + f[m+1]) - 0.0625*(f[m-1] + f[m+2]);
 }

/* ----------  end of mini-functions  ---------- */


/* ===================================================================== */


double x2y( double xi, double x0, double dx0, struct spectdata *sdata ) {
	double  w, ww, s, y=0.0, en;
	int    n;

	w = ( xi - x0 ) / dx0;
	ww = w + 0.5;
	en = floor( ww );
	s = w - en;
	n = (int) en;  
	if( n >= 0 && n <= sdata->ufile->num_elems ) { //FIXME in theory this check is no longer needed
		y = sdata->ufile->data[n] + ( s * ( sdata->coeff_b[n] + ( s * sdata->coeff_c[n] ) ) );
	}	
	return( y );
}

