#include <math.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* ========-======== function runthru() ======== */
/* 01 Jun 2003: Modified by Michael Koppelman to use fopen, fread, etc.
 * plus other minor formating changes for readability 
 */
/* 31 Oct 1992:  Function "runthru()" checks the contents of
   a uniformly-spaced spectrum file of the type produced by
   "unifspec.c".   (This function is considerably different
   from an earlier, Apr'92 version.)

   This is the TurboC version;  for differences between
   TurboC and unix search this file for the string "TurboC".

   The input file must consist of a header and data as follows:

   --(1) Text header, less than BUFFSIZE bytes long.  The end of the 
   header is the end of a text line containing the character pair #^
   (tictactoe and carat);  see explanatory note later below.*
   The header must include a parameter-spec string 
               #WVL1,DWVL=  <wvl1>  <dwvl>
   where <wvl1> = central wavelength of the first "wavelength 
   bin" and <dwvl> = width of wavelength bin = interval between
   successive samples, both to be read as floating point numbers.  
   (Note that the label "#WVLS,DWFL=" contains no spaces.)  
   The text header may have any length, and the "#WVL1,DWVL=..." 
   string may occur anywhere within it.

   --(2) The data, fluxes in binary 4-byte floating point (fp4)
   format, begin immediately after the #^ character pair which 
   terminates the text header.   The file must end at the last 
   byte of the data, i.e., no "trailer" or appended bytes
   are allowed.

     *Explanatory note to clarify how the text header is terminated:
     In order to be able to write a text header with vi, it should
     end with the character decimal 10 (LF), but the possible
     presence of a 13 (CR) just before the 10 should not matter
     (because this differs between unix and ms-dos forms of vi).
     A convenient character pair "#^" is used to signal the 
     last line of the header.  Thus the end of the header is
     identified in the following way:   As the header is read, this 
     function looks for the #^ signal.  When #^ is encountered,
     then any following bytes are read and discarded until
     a 10 (LF) character has been found.  Then the binary data
     are expected to begin with the next byte following that 
     10 character.

   

   ARGUMENTS OF FUNCTION "runthru()":

   filen[] (used as input to function):  file name, a character string 
     which is not altered by the function. 

   length[] (output only):  three 8-byte integers describing length of file;
        length[0] = length of text header in bytes,
        length[1] = number of data entries or "wavelength bins"
                  = 1/4 number of bytes in data section
        length[2] = total number of bytes found in file;  if this
             is larger than length[0] + 4*length[1], then probably
             something is wrong with some data values(?)
      The length[] parameters are found by the function;  the
      function does not use them as input arguments. 

   stats[] (output only):  four 8-byte f.p.'s as follows:
        stats[0] = value of WVL1 (first pixel wavelength, see above),
        stats[1] = value of DWVL (data interval or "bin width"),
        stats[2] = minimum data value (flux) found,
        stats[3] = maximum data value (flux) found.
      The first two of these are found in the text header (see above)
      and the other two are found while reading through the file
      to see how long it is.  Like length[], the values that happen
      to be in array stats[] when the function is called are not used,
      and are replaced by the values found as described above.

   RETURN VALUE should normally be 0.  The possibilities are 
      -100  if the file could not be opened
       -90  if the header could not be read
       -80  if no end-of-header could be found 
       -70  if the "#WVL1,DWVL=" tag was not found
       -60  if "#WVL1,DWVL=" was found but not the 2 numbers that
                 should follow it
       -50  if DWVL was found to be negative or zero
       -40  if the file could not be opened a second time, to read
                 the data*
       -30  if the header could not be read (skipped) a second time*
       -20  if no data values were found, following the header
         0  "normal":  Header was read successfully, WVL1 and DWVL
               were found, and at least one 4-byte binary data value 
               was found following the header.
    (* Return values -40 and -30 should never occur, even with a bad 
       input file.)
     

====-====  end of explanation notes; source code follows  ====
*/

#define BUFFSIZE	4096

int runthru(char *filen, long int *length, double *stats)   {
	FILE *fd;
	char tag[30];
	int i,j,m,n,o=0,returnval;
	int mk;
	long int bufsize,  fsize = 0L;
	long int nn; /* 8-byte int = long int for TurboC, int for unix? */
	double a;
	union { char header[BUFFSIZE];  float data[1024]; } buf;
    /* The maximum allowable header length here is BUFFSIZE bytes;  the total
       number of data entries following the header can be arbitrarily large,
       because 1024 is merely the number of data entries in one "read".  
       If size of buf is changed, then revise variable bufsize value below, 
       as well as the maximum header length mentioned in the notes above. */

	bufsize = BUFFSIZE;   /* corresponds to dimension declared above */

	length[0] = 0;
	length[1] = 0;
	length[2] = 0;
	stats[0] = 0.;
	stats[1] = 0.; 
	stats[2] = 1.e50;
	stats[3] = -stats[2];
    /* (4-byte f.p.'s overflow at 2**128 = 3.40282e38) */
	sprintf(tag,"#WVL1,DWVL=");

	returnval = 0;

	/* ====-==== Open the file and read the header, then close the file: */
	if ( (fd = fopen(filen,"r") ) == NULL ) {
		returnval = -100; 
     	/* returnval = -100 means "file can't be opened" */
	}

	if (returnval >= 0) {
		if( fseek( fd, fsize, SEEK_END ) == 0 ) {
			fsize = ftell( fd );
			if( fsize < bufsize )
				bufsize = fsize;
		}
		rewind( fd );
		
  		o = fread( &buf.header[0], bufsize, 1, fd );
		o = bufsize * o;
  		if (o <= 0) {
			returnval = -90;
       		/* returnval = -90 means "can't read the file" */
		}
 	}
  	fclose(fd);

	/* ====-==== If it's OK so far, look for the end of the header in the
             array that has just been read from the file: */
	if (returnval >= 0) {
		n = 0;
		for (i=0;i<(o-2);i++) {
   	 		if (buf.header[i] == '#' && buf.header[i+1] == '^') {
   	   			/* look for LF (char 10) following the #^ last-line marker: */ 
   	   			for (j=i+2;j<o;j++) {
   	     			if (buf.header[j] == 10) {  /* char = 10 is LF */
   	     				n = j + 1;
						break;
					}
   	    		}  /* end of j-loop  */
   	  		}  /* end of if char pair  == #^  */
   	 		if (n)
				break;
 		}  /* end of i-loop  */
  		if (n)
			length[0] = n;  
  		else {
			returnval = -80; 
  			/* returnval = -80 means "can't find end of header"  */
		}
 	} 	 /* ---- End of looking-for-end-of-header ---- */ 
	

	/* 	====-==== If it's OK so far, look for "#WVL1,DWVL=...":  */
	if (returnval >= 0) {
		returnval = -70;
   	  	/* If returnval is still -70 after the following procedure,
   	     it means "can't find tag '#WVL1,DWVL=' "  */
  		for (i=0;i<n-14;i++) {
   	 		if (buf.header[i] == tag[0]) {  /* '#' character */
   	   			returnval = -60;
   	   			for (j=1;j<=10;j++) {
   	   	  			if (buf.header[i+j] != tag[j]) {
							returnval = -70;
							break;
					}
   	    		} /* end of j-loop */
   	  		}
   	 		if (returnval > -70) {  /* have found #WVL1,DWVL= tag... */ 
   	   			stats[1] = 1.;
   	   			if (sscanf(&buf.header[i+11],"%le%le",&stats[0],&stats[1]) == 2)
   	        		returnval = 0;   /* have found WVL1 and DWVL */
   	   			else {
					returnval = -60;
   	     			/* returnval = -60 means "tag '#WVL1,DWVL=' was found, but
   	        		the two numbers WVL1 and DWVL were not found"  */
				}
   	   			if (stats[1] <= 0.) {
					returnval = -50;
   	     			/* returnval = -50 means "DWVL was zero or negative" */
				}
   	   			break;
   	  		}
   		} /* end of i-loop */
	 } 	 /* ---- End of looking-for-WVL1,DWVL in header ---- */

	
	/*  If everything is OK so far, re-open the file to read the data numbers 
     following the header:  */

	if (returnval >= 0) {
		if ( (fd = fopen(filen,"r") ) == NULL ) {
			returnval = -40; 
       		/* returnval = -40 means "file couldn't be opened second time" */
 		}

		if (returnval >= 0) {
  			/* First read (skip) the header: */
  			n = length[0];   
  			o = fread( &buf.header[0], n, 1, fd);
			o = n * o;
  			if (o != n) {
				returnval = -30;
     			/* returnval = -30 means "couldn't read through header second time" */
			}
		}

		/* If everything is OK so far, read the data following the header: */
		if (returnval >= 0) {
  			nn = 0;  /* nn will be number of bytes after the header */
  			n = 4;
			mk = bufsize/4;
  			for (;;) {
    			o = fread( &buf.data[0], n, mk, fd );
				o = n * o;
    			if (o < 0)
					o = 0;
    			nn = nn + o;
    			m = o/4;
    			for (i=0;i<m;i++) {
      				a = buf.data[i];
      				/* If abs value is very large, this indicates "invalid data": */
      				if (fabs(a) < 1.e36) {
       	 				if (a < stats[2]) {
							stats[2] = a;
						}
       	 				if (a > stats[3]) {
							stats[3] = a;
						}
       				}
     			}  /* end of i-loop */
    			if ( o < (n*mk) )
					break; 
   			} /* --- end of ;; loop */
  			length[1] = nn/4;
  			length[2] = length[0] + nn;
  			if (nn < 4) {
				returnval = -20;
     			/* returnval = -20 means "no data values were found after the header" */ 
			}
 		}
		fclose(fd);  
	}
	return returnval;
}
