
/* Function tickmks(), copied below, chooses several good places to put tick 
   marks.  Program main() is merely a simple routine to test tickmks(). */

#include <math.h>

/* Function tickmks(), July 2003: Given an interval xmin <= x <= xmax,
   this routine chooses 3 to 5 good places to put tick marks whose
   values of x are nearly as rounded as possible.  

   Input parameters:  xmin, xmax.   Note however, that if these are 
   reversed and xmax < xmin, then the program will nevertheless
   give an OK set of tick marks between the two limits. 

   Output values:  xtick[i] = x-values for tick marks, 0 <= i < Nticks.   
   Array xtick[] must be declared to have at least 8 elements
   ("double xtick[8]").   

   Return value:  Nticks, the number of valid tick marks, usually 3-5.  
   If the return value is 0, then something's wrong with xmin,xmax. 
   Note that the program refuses to deal with very small or very
   large x's:  it returns 0 if |xmax - xmin| is smaller than 1e-30 
   or larger than 1e+30. 

   This routine requires <math.h>.

   ------------------------ */

int tickmks(double xmin, double xmax, double * xtick)
 {
  int i,k,m,nticks[3];
  int good[7]={5, 2, 0, 1, 3, 4, 6}; 
    /* (Array good[] specifies desirability-priorities for
        1, 2, ... 7  tick marks (e.s., 3 ticks is best.) */
  double a,x1,x2,width,fac,dx[3],x0[3];

  for (m=0;m<8;m++) xtick[m] = 0.; 

  if (xmax >= xmin) { x1 = xmin;  x2 = xmax; }  /* normal */
  else { x1 = xmax;  x2 = xmin; }              /* reversed */
  width = x2 - x1;
  if (width < 1.e-30 || width > 1.e30) { m = 0;  nticks[0] = 0; } 
  else  /* normal case */
   {
     /* x1, x2 will be very close to xmin, xmax but we'll imperceptibly 
        widen the interval between them, in order to provide for tick
        marks that are practically at the ends of the range.  */
    a = 0.0001*width;
    x1 = x1 - a;
    x2 = x2 + a; 
    width = x2 - x1;  /* note that this is definitely > 0. */

    /* try three possible intervals dx:  0.5, 1.0, and 2.0 times 10^n */
    fac = 0.25; 
    for (k=0;k<3;k++)  
     {
      fac = 2.*fac;
      a = floor(log10(0.5*width/fac));  
        /* floor = largest integer that is not greater than the argument */

      dx[k] = fac*pow(10., a);   /* let's try this tick-interval */

      x0[k] = dx[k]*ceil(x1/dx[k]);   
        /* ceil = smallest integer that is not smaller than the argument: */ 
      nticks[k] = floor((x2 - x0[k])/dx[k]);  /* (number of ticks) - 1 */ 
      if (nticks[k] > 6) nticks[k] = 6;  
      if (nticks[k] < 0) nticks[k] = 0;  
     }
   }

  /* choose the k which gives best number of ticks: */ 
  m = 0;  i = good[nticks[0]]; 
  k = good[nticks[1]];
  if (k <= i) { m = 1;  i = k; }
  if (good[nticks[2]] <= i) m = 2;

   /* Next line is for an unusual case where we're using
      TurboC/MSDOS and the procedure failed because 2-byte 
      integers were used.  In principle, even 4-byte integers
      might fail if x2 is absurdly close to x1. */
  if (x0[m] < x1 || x0[m] > x2) nticks[m] = 0;
  else  /* calculate the valid tick marks: */
   {
    ++nticks[m];
    xtick[0] = x0[m];
    for (k=1;k<nticks[m];k++)  xtick[k] = xtick[k-1] + dx[m];
   }

  return nticks[m];  

 }  /*  -------- end of function tickmks() -------- */ 

/* ---------------------------------===-===-------------------------------- */

